@@ -65,6 +65,11 @@ private:
     void announceStop();
     void joinWorkers();
 
+    bool hasErrored();
+    bool allJobsDone();
+    bool waitForWakeUpEvent();
+    void rethrowExceptions();
+    
     std::vector<std::thread> workers_;        // worker threads in the pool
     std::queue<std::function<void()>> jobs_;  // the task que
 
@@ -74,6 +79,7 @@ private:
     std::condition_variable cvBusy_;
     size_t numBusy_{0};
     bool stopped_{false};
+    std::exception_ptr error_ptr_;
 };
 
 //! constructs a thread pool with as many workers as there are cores.
@@ -111,20 +117,17 @@ template<class F, class... Args>
 void ThreadPool::push(F&& f, Args&&... args)
 {
     if (workers_.size() == 0) {
-        // if there are no workers, just do the job in the main thread
-        f(args...);
+        f(args...); // if there are no workers, do the job in the main thread
+        return;
     } else {
-        // add job to the queue; must hold lock while modifying the shared queue
-        {
-            std::lock_guard<std::mutex> lk(mTasks_);
-            if (stopped_)
-                throw std::runtime_error("cannot push to joined thread pool");
-            jobs_.emplace([f, args...] { f(args...); });
-        }
-
-        // signal a waiting worker that there's a new job
-        cvTasks_.notify_one();
+        // must hold lock while modifying the shared queue
+        std::lock_guard<std::mutex> lk(mTasks_);
+        if (stopped_)
+            throw std::runtime_error("cannot push to joined thread pool");
+        jobs_.emplace([f, args...] { f(args...); });
     }
+    // signal a waiting worker that there's a new job
+    cvTasks_.notify_one();
 }
 
 //! pushes jobs returning a value to the thread pool.
@@ -137,31 +140,12 @@ template<class F, class... Args>
 auto ThreadPool::pushReturn(F&& f, Args&&... args)
     -> std::future<decltype(f(args...))>
 {
-    using result_t = decltype(f(args...));
-    using jobPackage = std::packaged_task<result_t()>;
-
-    // create packaged task on the heap
+    using jobPackage = std::packaged_task<decltype(f(args...))()>;
     auto job = std::make_shared<jobPackage>([&f, args...] {
         return f(args...);
     });
+    this->push([job] { (*job)(); });
 
-    if (workers_.size() == 0) {
-        // if there are no workers, just do the job in the main thread
-        (*job)();
-    } else {
-        // add job to the queue
-        {
-            std::lock_guard<std::mutex> lk(mTasks_);
-            if (stopped_)
-                throw std::runtime_error("cannot push to joined thread pool");
-            jobs_.emplace([job] { (*job)(); });
-        }
-
-        // signal a waiting worker that there's a new job
-        cvTasks_.notify_one();
-    }
-
-    // return future result of the job
     return job->get_future();
 }
 
@@ -211,9 +195,8 @@ inline void ThreadPool::parallelFor(ptrdiff_t begin, ptrdiff_t end,
         for (ptrdiff_t i = b.begin; i < b.end; i++) f(i);
     };
     auto batches = createBatches(begin, end - begin, workers_.size(), nBatches);
-    for (const auto& batch : batches) {
+    for (const auto& batch : batches)
         this->push(doBatch, batch);
-    }
 }
 
 //! computes a for-each loop in parallel batches.
@@ -254,39 +237,21 @@ inline void ThreadPool::parallelForEach(I& items, F&& f, size_t nBatches)
 //! but does not join the threads.
 inline void ThreadPool::wait()
 {
-    if (workers_.size() == 0) {
-        // all jobs have been executed in the main thread, no need to wait
-        checkUserInterrupt();
-        Rcout << "";
-        return;
-    }
-
-    auto allJobsDone = [this] { return (numBusy_ == 0) && jobs_.empty(); };
-    auto timeout = std::chrono::milliseconds(250);
     while (true) {
-        {
-            // wait_for tries acquires lk when waking up
-            std::unique_lock<std::mutex> lk(mTasks_);
-            cvBusy_.wait_for(lk, timeout, allJobsDone);
-            // lk can be released immediately
+        if ( waitForWakeUpEvent() ) {
+            if ( !this->allJobsDone() ) {
+                this->clear();  // cancel all remaining jobs
+                continue;       // wait for currently running jobs
+            }
+            if ( this->hasErrored() | this->allJobsDone() | isInterrupted() )
+                break;
         }
-
-        // check whether timeout was reached or another event caused wake-up
-        if ( isInterrupted() ) {
-            this->clear();  // cancel all remaining jobs
-            break;
-        } else if ( allJobsDone() ) {
-            break;
-        }
-
-        // give other threads priority before waiting again
         Rcout << "";
         std::this_thread::yield();
     }
 
-    // synchronize one last time before continuing main thread
     Rcout << "";
-    checkUserInterrupt();
+    this->rethrowExceptions();
 }
 
 //! waits for all jobs to finish and joins all threads.
@@ -329,9 +294,13 @@ inline void ThreadPool::startWorker()
             job = std::move(jobs_.front());
             jobs_.pop();
 
-            // lock can be released before starting work
+            // lock can be released before starting work, but must signal
+            // that thread will be busy before (!) to avoid premature breaks
+            this->announceBusy();
             lk.unlock();
+
             this->doJob(std::move(job));
+            this->announceIdle();
             std::this_thread::yield();
         }
     });
@@ -341,27 +310,19 @@ inline void ThreadPool::startWorker()
 //! @param job job to be exectued.
 inline void ThreadPool::doJob(std::function<void()>&& job)
 {
-    checkUserInterrupt();
-    this->announceBusy();
+    // checkUserInterrupt();
     try {
         job();
-    } catch (const std::exception& e) {
-        this->announceIdle();
-        throw e;
     } catch (...) {
-        this->announceIdle();
-        throw std::runtime_error("caught unknown C++ exception.");
+        std::lock_guard<std::mutex> lk(mTasks_);
+        error_ptr_ = std::current_exception();
     }
-    this->announceIdle();
 }
 
-//! signals that a worker is busy.
+//! signals that a worker is busy (must be called why locking mTasks).
 inline void ThreadPool::announceBusy()
 {
-    {
-        std::lock_guard<std::mutex> lk(mTasks_);
-        ++numBusy_;
-    }
+    ++numBusy_;
     cvBusy_.notify_one();
 }
 
@@ -396,5 +357,37 @@ inline void ThreadPool::joinWorkers()
     }
 }
 
+//! checks if an error occured.
+inline bool ThreadPool::hasErrored()
+{
+    return static_cast<bool>(error_ptr_);
+}
+
+//! check whether all jobs are done
+inline bool ThreadPool::allJobsDone()
+{
+    return (numBusy_ == 0) && jobs_.empty();
+}
+
+//! checks whether wait() needs to wake up
+inline bool ThreadPool::waitForWakeUpEvent()
+{
+    static auto timeout = std::chrono::milliseconds(250);
+    auto isWakeUpEvent = [this] {
+        return this->allJobsDone() | this->hasErrored();
+    };
+    std::unique_lock<std::mutex> lk(mTasks_);
+    cvBusy_.wait_for(lk, timeout, isWakeUpEvent);
+    return isWakeUpEvent();
+}
+
+//! rethrows exceptions (exceptions from workers are caught and stored; the
+//! wait loop only checks, but does not throw for interruptions) 
+inline void ThreadPool::rethrowExceptions()
+{
+    checkUserInterrupt();
+    if ( this->hasErrored() )
+        std::rethrow_exception(error_ptr_);
+}
 
 }
